@@ -30,6 +30,7 @@
 #define NM_MANAGER_WWAN_HARDWARE_ENABLED        "wwan-hardware-enabled"
 #define NM_MANAGER_WIMAX_ENABLED                "wimax-enabled"
 #define NM_MANAGER_WIMAX_HARDWARE_ENABLED       "wimax-hardware-enabled"
+#define NM_MANAGER_RADIO_FLAGS                  "radio-flags"
 #define NM_MANAGER_ACTIVE_CONNECTIONS           "active-connections"
 #define NM_MANAGER_CONNECTIVITY                 "connectivity"
 #define NM_MANAGER_CONNECTIVITY_CHECK_AVAILABLE "connectivity-check-available"
@@ -199,6 +200,10 @@ NMMetered nm_manager_get_metered(NMManager *self);
 
 void nm_manager_notify_device_availability_maybe_changed(NMManager *self);
 
+struct _NMDnsManager;
+
+struct _NMDnsManager *nm_manager_get_dns_manager(NMManager *self);
+
 /*****************************************************************************/
 
 void nm_manager_device_auth_request(NMManager                     *self,
