@@ -377,7 +377,7 @@ nmt_editor_constructed(GObject *object)
     else if (nm_connection_is_type(priv->edit_connection, NM_SETTING_WIREGUARD_SETTING_NAME))
         page = nmt_page_wireguard_new(priv->edit_connection, deventry);
     else
-        g_assert_not_reached();
+        g_return_if_reached();
 
     add_sections_for_page(editor, grid, page);
     nmt_editor_grid_append(grid, NULL, nmt_newt_separator_new(), NULL);
